/*=======================================================
  Version:  $Id: mod2kptrs.c,v 1.3 1999/02/01 19:59:07 muth Exp $
  Comment:  This file contains code for the mod 2 k alias analysis
=======================================================*/

/*
  mod2kptrs.c - implements MAY alias analysis using modular arithmetic
  has a hash function.  The main external functions are noConflict, which
  requires that ComputeMod2kptrs be called before use.

  NOTES:

  At each program point before and after a basic block I maintain a
  mapping from a register to an address descriptor which I call an
  mset.  This mapping is maintained in a summary which is just
  typedef'd to be an array of mset's.  IN[b] is the summary of the
  address descriptors before the basic block b.

  **** ALL THESE NOTES (below) ARE OUTDATED ****

  function noConflict returns true if it can be proven
  that a memory access in instruction i1, definitely does not conflict
  with with memory access i2.  i.e. i1 and i2 access different
  locations in memory

  mset is a set of modular 2^k numbers plus a base value b, that
  indicate what values a register may contain at a given point the
  CFG.

  At a specific program point, consider the value v in register r,
  with mset M.  It must be that v in { i*2^k + m + M.b | for all m in
  M.set }.

  b is represented as a pointer to an instruction that defines an
  unknown value for its destination register.  For example a load from
  memory.

  The base value for two mset's are equal for two different
  instructions only if i0 Dom i1 Dom i2.  Where i0 is the base value
  of *BOTH* msets.

  #defines -

  STACK_TRICK - tries to mimick roberts stack trick.  Assumes that
  stack pointer can never be evaluated in program.  Therefore if one
  instruction can be evaluated and the other is relative to the stack
  pointer then the address must not conflict.

  LOCAL_STACK - sets the base pointer for every init block that
  defines the stack pointer in hopes to keep propagated base pointers
  to point to the local stack frame.

  MOD2K_DEBUG - prints out some crap to stderr for debugging.

  QUESTIONS:

  TODO:
  o what about accesses of different sizes.
  
  */

/***********************************************************************
  headers 
  *********************************************************************/
#include "global.h"
#include "functions.h"
#include "flags.h"
#include "opcode.h"
#include "regs.h"

#include "mod2kptrs.h"
#include "mset.h"

#ifdef MOD2K_DEBUG
#include <sys/timers.h>
int getclock(int clktyp, struct timespec *tp);
struct timespec start_time, end_time;
double time_start, time_end;
#endif

#define STACK_TRICK
#define AINS_IS_UNALIGNED(i) (AINS_CODE(i) == I_stq_u || AINS_CODE(i) == I_ldq_u)

/***********************************************************************
  Prototypes
  *********************************************************************/
static void summary_empty(bbl_summary s);
static void summary_all(bbl_summary s);
static BOOL summary_copy(bbl_summary s1, bbl_summary s2);
static BOOL summary_union_to(bbl_summary s1, bbl_summary s2);
void summary_eval(INDEX i, bbl_summary OUT, bbl_summary IN);
BOOL insSameWidth(INDEX i1, INDEX i2);
#ifdef MOD2K_DEBUG
static void summary_printem(FILE *fp, bbl_summary s);
#endif

/***********************************************************************
  Globals
  *********************************************************************/
bbl_summary *IN = NULL;
bbl_summary *OUT = NULL;

ins_summary *addrDesc = NULL;

#undef __PROC__
#define __PROC__ "noConflict"

/* noConflict() returns TRUE only if the address accessed by instruction 
   i1 is provably different from the address accessed by instruction i2 */   

BOOL noConflict(INDEX i1, INDEX i2)
{
  mset m1, m2, tmp;
  int i;

  if (! ((AINS_IS_LOAD(i1) || AINS_IS_STORE(i1)) &&
	 (AINS_IS_LOAD(i2) || AINS_IS_STORE(i2)))) {
#if 0
    CASSERT(((AINS_IS_LOAD(i1) || AINS_IS_STORE(i1)) &&
	     (AINS_IS_LOAD(i2) || AINS_IS_STORE(i2))),
	    ("instr not load or store:\n"));
    if (!(AINS_IS_LOAD(i1) || AINS_IS_STORE(i1))) PrIns(i1);
    else PrIns(i2);
    printf("\n");
#endif
    return FALSE;
  }

  if (!addrDesc) {
    CWARNING(addrDesc, ("addrDesc not computed: Need to run ComputeMod2kptrs?\n"));
    return FALSE;
  }
  /*
   * For an address expression of the form k(Rb) the value of k for 
   * instruction i is in AINS_DATA(i).
   */
  mset_addi(&m1, &addrDesc[i1][Rb], AINS_DATA(i1), i1);
  mset_addi(&m2, &addrDesc[i2][Rb], AINS_DATA(i2), i2);

  /* to deal with multibyte memory accesses, have the mset indicate
     that the register may point to any of the bytes that it may
     access */
  mset_copy(&tmp, &m1);
  for (i = 0; i < AinsMemWidth(i1) - 1; i++) {
    mset_addi(&tmp, &tmp, 1, i1);
    mset_union(&m1, &tmp, &m1);
  }
  
  mset_copy(&tmp, &m2);
  for (i = 0; i < AinsMemWidth(i2) - 1; i++) {
    mset_addi(&tmp, &tmp, 1, i2);
    mset_union(&m2, &tmp, &m2);
  }
  
  if (!mset_sameValue(&m1, i1, &m2, i2)) {
    /* m1 and m2 are defined in terms of a possibly different base
       values */

#ifdef STACK_TRICK
    if (m1.basep == BASE_TOP || m2.basep == BASE_TOP)
      return FALSE;
    
    if ((VALID_BASE(m1.basep) && AINS_REG_C(m1.basep) == REG_SP && !VALID_BASE(m2.basep)) ||
	(VALID_BASE(m2.basep) && AINS_REG_C(m2.basep) == REG_SP && !VALID_BASE(m1.basep)))
      /* one of the msets is absolute and the other is relative to the
         local stack frame */
      return TRUE;
#endif
    	
    return FALSE;
  }
    
  /* there is no conflict only if the two msets are disjoint */
  return mset_disjoint(&m1, &m2);
} /* noConflict */ 

#undef __PROC__
#define __PROC__ "ComputeMod2kptrs"

extern void ComputeMod2kptrs()
{
  INDEX b, p;
  INDEX i;
  int counter = 0;
  bbl_summary tmp;
  BOOL changed;
  
#ifdef MOD2K_DEBUG
  getclock(TIMEOFDAY, &start_time);
#endif

  /* allocate space 
  printf("\tallocating 2x%dx%d= %d bytes\n",
	 2*CONFIG_UBBL, (int) sizeof(bbl_summary),
	 2*CONFIG_UBBL* (int) sizeof(bbl_summary)); */
  
  IN = calloc(2*CONFIG_UBBL, sizeof(bbl_summary));
  if (addrDesc)
    free(addrDesc);
  
  addrDesc = calloc(CONFIG_UINS, sizeof(ins_summary));
  CASSERT(IN && addrDesc, ("Cannot allocate 2x%dx%d = %d bytes\n",
	       2*CONFIG_UBBL, sizeof(bbl_summary), 2*CONFIG_UBBL*sizeof(bbl_summary)));
  OUT = &IN[CONFIG_UBBL];
 
  /* initialize OUT to be the gen of each BBL
     and the IN to be empty */
  FOREACH_LIVE_BBL(b) {
    summary_empty(IN[b]);
    summary_empty(OUT[b]);

    /* imprecision in CFG - NOW HANDLED BY AFUN_HELL
       if (ABBL_IS_FIRST(b) && ABBL_IS_RELOC(b))
       summary_all(IN[b]); */

#ifdef LOCAL_STACK
    /* make the stack pointer an exception...  if the first block of a
       function has an instruction with destination SP as the
       register, then set the base pointer to be this instruction.
       This insures that if a ptr is on the stack frame its mset basep
       will point to the instruction defining the current stack frame
       rather than some other stack frame. */       
    if (ABBL_IS_FIRST(b)) {
      FOREACH_BBL_INS(b, i) {
	if (AINS_REG_C(i) == REG_SP)
	  mset_base(&OUT[b][REG_SP], i);
      }
    }
#endif
    summary_bblgen(OUT[b], IN[b], ABBL_INS_LAST(b));
  }

#if defined(MOD2K_DEBUG) && !defined(USE_BASEP)
  {
    int i;

    FOREACH_LIVE_BBL(b)
      for (i = 0; i < NREGS; i++)
	if (IN[b][i].basep > 0) {
	  fprintf(stderr, "IN for BBL: %d\n", b);
	  summary_printem(stderr, IN[b]);
	  fprintf(stderr, "OUT for BBL: %d\n", b);
	  summary_printem(stderr, OUT[b]);
	  break;
	}
  }
#endif
  /* imprecision in CFG */

  FOREACH_LIVE_BBL(b)
  {
      if( AFUN_IS_PSEUDO(ABBL_FUN(b)) )
          summary_all(OUT[b]);
  }

  /* initialize fixpoint computation */
  FOREACH_LIVE_BBL(b) {
    ABBL_FLAGS(b) |= F_MARKED;
  }

  /* do a fixpoint computation */
  do {
    counter++;
    changed = FALSE;
    FOREACH_LIVE_BBL(b) {
      if ( (ABBL_FLAGS(b) & F_MARKED) && !AFUN_IS_PSEUDO(ABBL_FUN(b)) )
      {      
	ABBL_FLAGS(b) &= ~F_MARKED;
	summary_empty(tmp);
	FOREACH_BBL_PRED(b, p) {  /* IN[b] = union{OUT[p] | p a predecessor of B} */
	  summary_union_to(tmp, OUT[AEDG_FROM(p)]); 
	}	
	if (summary_copy(IN[b], tmp)) {
	  /* change in input could be "undone" by bbl, recompute changed */
	  changed = summary_bblgen(OUT[b], IN[b], ABBL_INS_LAST(b)) || changed;
	}
	if (changed) {  /* mark successor blocks */
	  FOREACH_BBL_SUCC(b, p) {
	      ABBL_FLAGS(AEDG_TO(p)) |= F_MARKED;
	  }
	}
      }
    } /* FOREACH_LIVE_BBL */
  } while (changed);
 
  /* convert to instruction summary */
  free(OUT);
  FOREACH_LIVE_BBL(b) {
    summary_copy(tmp, IN[b]);
    FOREACH_BBL_INS(b, i) {
      if (!REG_IS_INT(AINS_REG_A(i)))
	mset_empty(&addrDesc[i][Ra]);
      else
	mset_copy(&addrDesc[i][Ra], &tmp[AINS_REG_A(i)]);
      
      if (!REG_IS_INT(AINS_REG_B(i)))
	mset_empty(&addrDesc[i][Rb]);
      else
	mset_copy(&addrDesc[i][Rb], &tmp[AINS_REG_B(i)]);
      
      if (!REG_IS_INT(AINS_REG_C(i)))
	mset_empty(&addrDesc[i][Rc]);
      else
	mset_copy(&addrDesc[i][Rc], &tmp[AINS_REG_C(i)]);
      
      summary_eval(i, tmp, tmp);
    } /* FOREACH_BBL_INS */
  } /* FOREACH_BBL_INS */

#ifdef MOD2K_DEBUG
  getclock(TIMEOFDAY, &end_time);
  time_start = start_time.tv_sec*1.0e9 + start_time.tv_nsec;
  time_end = end_time.tv_sec*1.0e9 + end_time.tv_nsec;
#endif

#ifdef MOD2K_DEBUG
  FOREACH_LIVE_INS(i) {
    CASSERT(addrDesc[i][Ra].basep < CONFIG_UINS,
	    ("basep out of range! ins = %d, bbl = %d, Ra = %d", i, AINS_BBL(i), AINS_REG_A(i)));
    CASSERT(addrDesc[i][Rb].basep < CONFIG_UINS,
	    ("basep out of range! ins = %d, bbl = %d, Rb = %d", i, AINS_BBL(i), AINS_REG_B(i)));
    CASSERT(addrDesc[i][Rc].basep < CONFIG_UINS,
	    ("basep out of range! ins = %d, bbl = %d, Rc = %d", i, AINS_BBL(i), AINS_REG_C(i)));
  }
#endif
  
  free(IN);

#ifdef MOD2K_DEBUG 
  {
    int nTotal[NREGS];
    int nEmpty[NREGS];
    int nAll[NREGS];
    int nOne[NREGS];
    int nFew[NREGS];
    int nBaseOne[NREGS];
    int nBaseFew[NREGS];
    int nBaseEmpty[NREGS];

    int sumTotal;
    int sumEmpty;
    int sumAll;
    int sumOne;
    int sumFew;
    int sumBaseOne;
    int sumBaseFew;
    int sumBaseEmpty;
    int r;
    mset_value_t m;
    int nM;
    int bbl = 0, instrs = 0;

    FOREACH_LIVE_INS(i) {
      if (REG_IS_INT(AINS_REG_A(i)) && addrDesc[i][Ra].basep == -1 &&
	  addrDesc[i][Ra].set == BITV_EMPTY)
	fprintf(stderr, "Instr %d in BBL: %d has empty set for register %d\n",
		i, AINS_BBL(i), AINS_REG_A(i));
      
      if (REG_IS_INT(AINS_REG_B(i)) &&
	  addrDesc[i][Rb].basep == -1 && addrDesc[i][Rb].set == BITV_EMPTY)
	fprintf(stderr, "Instr %d in BBL: %d has empty set for register %d\n",
		i, AINS_BBL(i), AINS_REG_B(i));
      
      if (REG_IS_INT(AINS_REG_C(i)) &&
	  addrDesc[i][Rc].basep == -1 && addrDesc[i][Rc].set == BITV_EMPTY)
	fprintf(stderr, "Instr %d in BBL: %d has empty set for register %d\n",
		i, AINS_BBL(i), AINS_REG_C(i));
    } /* FOREACH_LIVE_INS */
    
    
    for (r = 0; r < NREGS; r++) {
      nTotal[r] =  nEmpty[r] = nAll[r] = nOne[r] = nFew[r] = 0;
      nBaseOne[r] = nBaseFew[r] = nBaseEmpty[r]= 0;
    }    
      
#define DO_STATS(i, a, r) 						\
    nTotal[r]++;							\
    if (addrDesc[i][a].basep == -2 || addrDesc[i][a].set == BITV_ALL)	\
      nAll[r]++;							\
    else if (addrDesc[i][a].basep == -1) {				\
      nM = 0;								\
      BITV_FOREACH(addrDesc[i][a].set, m) 				\
	nM++;								\
      if (nM == 0)							\
	nEmpty[r]++;							\
      else if (nM == 1)							\
	nOne[r]++;							\
      else								\
	nFew[r]++;							\
    }									\
    else {								\
         nM = 0;							\
         BITV_FOREACH(addrDesc[i][a].set, m) 				\
            nM++;							\
         if (nM == 0)							\
	    nBaseEmpty[r]++;						\
         else if (nM == 1)						\
	    nBaseOne[r]++;						\
         else								\
	    nBaseFew[r]++;						\
    }

#define DO_SUM							\
    sumTotal = sumEmpty =  sumAll = sumOne = 0;			\
    sumFew = sumBaseOne = sumBaseFew = sumBaseEmpty = 0;	\
    for (r = 0; r < NREGS; r++) {				\
      sumTotal += nTotal[r];					\
      sumEmpty += nEmpty[r];					\
      sumAll += nAll[r];					\
      sumOne += nOne[r];					\
      sumFew += nFew[r];					\
      sumBaseOne += nBaseOne[r];				\
      sumBaseFew += nBaseFew[r];				\
      sumBaseEmpty += nBaseEmpty[r];				\
    }

	   
    /* statistics */
    FOREACH_LIVE_BBL(b) {
      FOREACH_BBL_INS(b, i) {
	if (REG_IS_INT(AINS_REG_A(i)))
	  DO_STATS(i, Ra, AINS_REG_A(i));

	if (REG_IS_INT(AINS_REG_B(i)))
	  DO_STATS(i, Rb, AINS_REG_B(i));
	  
	if (REG_IS_INT(AINS_REG_C(i)))
	  DO_STATS(i, Rc, AINS_REG_C(i));
      }
    }

    DO_SUM;
    for (r = 0; r < NREGS; r++) {
      nTotal[r] =  nEmpty[r] = nAll[r] = nOne[r] = nFew[r] = 0;
      nBaseOne[r] = nBaseFew[r] = nBaseEmpty[r]= 0;
    }    

    FOREACH_LIVE_BBL(b) {
      FOREACH_BBL_INS(b, i) {
	if (AINS_IS_LOAD(i) || AINS_IS_STORE(i)) {
	  CASSERT(REG_IS_INT(AINS_REG_B(i)), ("load with non integer register Rb"));
	  DO_STATS(i, Rb, AINS_REG_B(i));
	} /* if */
      } /* FOREACH_BBL_INS */
    } /* FOREACH_LIVE_BBL */
	      
    DO_SUM;

    printf("\nMemInstrs: Total %9d ", sumTotal);
    printf("  Unknown %d [ %6.3f ] ", sumAll, sumTotal ? sumAll*100.0/sumTotal : 0.0);
    printf("  b+one %d [ %6.3f ] ", sumBaseOne, sumTotal ? sumBaseOne*100.0/sumTotal : 0.0);
    printf("  b+few %d [ %6.3f ]\n", sumBaseFew, sumTotal ? sumBaseFew*100.0/sumTotal : 0.0);

     FOREACH_LIVE_BBL(b) bbl++;

     FOREACH_LIVE_INS(b) instrs++;

     printf("SIZE: bbls: %d   instrs: %d   time: %.2f sec\n", bbl, instrs, (time_end-time_start)/1.0e9);
  }
#endif

} /* ComputeMod2kptrs */

  /*
    char *answer[] = {"no", "maybe", "yes" };
    ComputeMod2kptrs();
    PrintIns(stderr, I1);
    PrintIns(stderr, I2);
    fprintf(stderr, "modk says %s, intra block says %s\n",
    noConflict(I1, I2) ? "no" : "maybe",
    answer[MemoryConflict(I1, I2)+1]);
    */
    
void summary_eval(INDEX i, bbl_summary OUT, bbl_summary IN)
{
  mset tmp;
  
  if (AINS_REG_C(i) == REG_ZERO)
    /* keep zero a constant */
    return;
  
  if (!REG_IS_INT(AINS_REG_C(i)))
    /* i don't care about floating point registers */
    return;
  
  CASSERT(AINS_REG_A(i) != REG_LIT, ("instr with literal rega\n"));
  
  if (AINS_FLAGS(i)&FI_EVAL_C) {
    /* we evaluated this instruction so just store the
       value in the set */
    mset_singleton(&OUT[AINS_REG_C(i)], AINS_C_EVAL(i));
    return;
  }

  switch(AINS_TYPE(i)) {
  case IT_X10:
  case IT_X11:
  case IT_X12:
  case IT_X13:
  case IT_X14:
  case IT_X15:
  case IT_X16:
  case IT_X17:
  case IT_X18:
  case IT_X1A:
  case IT_X1C:
  case IT_ERR:    /* flag error */
    WARNING(("instrcution type is FMT_ERR or FMT_X1?\n"));
    break;
    
  case IT_F2I:    /* move float reg to int reg */
  case IT_RX:
  case IT_MEM:
  case IT_MSC:    /* misc instructions */
  case IT_BRA:    /* uncoditional branch: br bsr */
  case IT_JMP:    /* jump */
  case IT_ILD:    /* integer load */
#ifdef USE_BASEP
    mset_base(&OUT[AINS_REG_C(i)], i);
#else
    mset_all(&OUT[AINS_REG_C(i)]);
#endif
    break;
      
  case IT_PAL:    /* pal is bad */
    summary_all(OUT);
    break;
    
  case IT_LDA:    /* load address */
    if (AINS_CODE(i) == I_ldah)
      return;
    
    mset_addi(&OUT[AINS_REG_C(i)], &IN[AINS_REG_B(i)], AINS_DATA(i), i);
    break;
    
  case IT_FCY:
  case IT_FM4:    /* float operations */
  case IT_FM8:    /* float operations */
  case IT_FD4:    /* float operations */
  case IT_FD8:    /* float operations */
  case IT_FOP:    /* float operations */
  case IT_I2F:    /* move int reg to float reg */
  case IT_FLD:    /* float load */
  case IT_FCM:    /* float conditional move */
  case IT_F2R:
  case IT_R2F:
    /* I don't care about floating point registers */	 
    break;
    
  case IT_IST:    /* integer store */
  case IT_FST:    /* float store */
  case IT_IBR:    /* integer cond branch */
  case IT_FBR:    /* float cond branch */
    /* these instruction don't define any registers */
    break;
    
  case IT_ICM:    /* integer conditional move */
    if (AINS_REG_B(i) == REG_LIT) {
      mset_singleton(&tmp, AINS_DATA(i));
      mset_union(&OUT[AINS_REG_C(i)], &tmp, &IN[AINS_REG_C(i)]);
    } /* if */
    else
      mset_union(&OUT[AINS_REG_C(i)], &IN[AINS_REG_C(i)], &IN[AINS_REG_B(i)]);
    break;
    
  case IT_LOG:
  case IT_SFT:
  case IT_MLL:
  case IT_MLQ:
  case IT_MLH:
  case IT_IOP:    /* integer operation */
    switch(AINS_CODE(i)) {
    default:
      /* PUNT (on all but simple arithmetic) */
      mset_base(&OUT[AINS_REG_C(i)], i);
      break;
      
      /* this macro checks to see if any of the source operands are
	 literals and calls the corresponding mset_operation op with
	 the correct arguments */
#define MOD_DOOP(op) 									\
      if (AINS_REG_A(i) == REG_LIT)							\
        mset_ ## op ## i(&OUT[AINS_REG_C(i)], &IN[AINS_REG_B(i)], AINS_DATA(i), i);	\
      else if (AINS_REG_B(i) == REG_LIT)						\
        mset_ ## op ## i(&OUT[AINS_REG_C(i)], &IN[AINS_REG_A(i)], AINS_DATA(i), i);	\
      else										\
        mset_ ## op (&OUT[AINS_REG_C(i)], &IN[AINS_REG_A(i)], &IN[AINS_REG_B(i)], i);

    case I_subl:
    case I_subl_v:
    case I_subq:
    case I_subq_v:
      MOD_DOOP(sub);
      break;
      
    case I_mull:
    case I_mulq:
    case I_mull_v:
    case I_mulq_v:
      MOD_DOOP(mul);
      break;

    case I_addq:
    case I_addq_v:
    case I_addl:
    case I_addl_v:
      MOD_DOOP(add);
      break;

      /* see MOD_DOOP above -  this does the scale as well */
#define MOD_DOOPS(op, scale) 						\
      mset_muli(&tmp, &IN[AINS_REG_A(i)], scale, i);			\
      if (tmp.basep == i)						\
	mset_copy(&OUT[AINS_REG_C(i)], &tmp);				\
      else if (AINS_REG_B(i) == REG_LIT)				\
        mset_ ## op ## i(&OUT[AINS_REG_C(i)], &tmp, AINS_DATA(i), i);	\
      else								\
        mset_ ## op (&OUT[AINS_REG_C(i)], &tmp, &IN[AINS_REG_B(i)], i);
	
    case I_s8addq:
    case I_s8addl:
      MOD_DOOPS(add, 8);
      break;

    case I_s8subq:
    case I_s8subl:
      MOD_DOOPS(sub, 8);
      break;
      
    case I_s4addq:
    case I_s4addl:
      MOD_DOOPS(add, 4);
      break;
      
    case I_s4subq:
    case I_s4subl:
      MOD_DOOPS(sub, 4);
      break;
      
    } /* switch on CODE */
  } /* switch on TYPE */
} /* summary_eval */

#undef __PROC__
#define __PROC__ "summary_bblgen"

BOOL summary_bblgen(bbl_summary out, bbl_summary in, INDEX lastIns) {
  INDEX b = AINS_BBL(lastIns);
  INDEX i;
  bbl_summary sum;

  summary_copy(sum, in);
  for (i = ABBL_INS_FIRST(b); VALID(i); i = AINS_NEXT(i)) {
    if (AINS_PREV(i) == lastIns)
      break;
    
    summary_eval(i, sum, sum);
  } /* FOREACH_BBL_INS */

  return summary_union_to(out, sum);
} /* summary_bblgen */


void summary_empty(bbl_summary s) {
  int i;

  for (i = 0; i < NREGS; i++)
    mset_empty(&s[i]);

  /* this should be the only assignment to REG_ZERO ever */
  mset_singleton(&s[REG_ZERO], 0);
} /* summary_empty */


BOOL summary_union_to(bbl_summary s1, bbl_summary s2) {
  int i;
  BOOL r = FALSE;

  for (i = 0; i < NREGS - 1; i++) 
    r = mset_union(&s1[i], &s1[i], &s2[i]) || r;

  return r;
} /* summary_union_to */


BOOL summary_copy(bbl_summary s1, bbl_summary s2) {
  int i;
  BOOL r = FALSE;

  for (i = 0; i < NREGS; i++) 
    r = mset_copy(&s1[i], &s2[i]) || r;

  return r;
} /* summary_copy */


void summary_all(bbl_summary s) {
  int i;

  for (i = 0; i < NREGS - 1; i++) 
    mset_all(&s[i]);

} /* summary_all */


#ifdef MOD2K_DEBUG
void summary_printem(FILE *fp, bbl_summary s)
{
  REG r;

  if (fp == NULL)
    fp = stderr;
  
  for (r = 0; r < NREGS; r++) {
    fprintf(fp, "register r%02d ", r);
    mset_printem(fp, &s[r]);
  } /* for */
} /* summary_printem */
#endif

