/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.switch.c,v 1.23 1999/04/20 21:24:51 muth Exp $
  Comment:  This file contains functions dealing with
            jump tables
=======================================================*/

#include "global.h"
#include "functions.h"

#include "flags.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
  The jump table idiom we are trying to detect here
  looks like this

  
  ZAP   [addl A,B,C] OPTIONAL
        [zapnot A,B,C] OPTIONAL
        
        A or B must be zero.
        This is usually a  zapnot instruction
        but the normalizer will take care of it.

  CMP   cmple  D,E,F  
        cmpult D,E,F       
        cmpule D,E,F

        C or D must be constant
        if ZAP is present C must agree with the
        non constant (D or E)
  
  IBR   beq    G,target         
        bne    G,target

        G must agree with F

-------------------------------------------------------------------------------
        
  ADJ   [addl  H,I,J]  OPTIONAL  
        [subl  H,I,J]  OPTIONAL    
        [zapnot H,I,J] OPTIONAL
        
        This could also be  zapnot instruction
        but the normalizer will take care of it.
        I or H must be constant.
        if ZAP is present the non constant (I or H) must agree
        with the non zero operand (A or B) other wise
        it must agree with the non constant operand of CMP (D or E)
        
  IND   s4addq K,L,M

        L must be constant and  pointing into a readonly section.
        if ADJ is present K must agree with J
        else if ZAP is present K must agree with the non zero operand
        of CMP (D or E).
  
  LDL   ldl    O,offset(P)

        O must agree with M
  
  ADD   addq   Q,R,S

        Q must agree with M
        R must be GP
        
  JMP   jmp    T

        T must agree with S
=======================================================*/




/*=======================================================
Typical table jump:

                            cmpult  $1,0x8,$0
                            beq     $0,true,false

000000012000ab78 120072ab0: ldq      $28,0xa050($29)    [___A_C]  88996
                 120072ab4: s4addq   $24,$28,$24        [____B_]  88997
                 120072ab8: ldl      $24,0x0000($24)    [___A__]  88998
                 120072abc: addq     $24,$29,$24        [______]  88999
0000000000000000 120072ac0: jmp      $31 ($24),0x000000 [___A_C]  89000

other:
                            cmpule  $1,0x8,$0
                            beq     $0,true,false

0000000120016220 120036380: ldq      $28,0x8870($29)    [___A_C]  27096
                 120036384: s4addq   $1 ,$28,$28        [____B_]  27097
                 120036388: ldl      $28,0x0658($28)    [___A__]  27098
                 12003638c: addq     $28,$29,$28        [______]  27099
0000000000000000 120036390: jmp      $31 ($28),0x000000 [___A_C]  27100

other:

                 120006760: cmple    $17,0x7a,$20       [____B_]    176
0000000000000000 120006768: beq      $20,t 81 f 42      [____BC]    178

0000000120000350 12000676c: ldq      $28,0x8218($29)    [___ABC]    179
                 120006770: lda      $17,0xffba($17)    [_M_A__]    180
                 120006774: s4addq   $17,$28,$28        [____B_]    181
                 120006778: ldl      $28,0x0264($28)    [___A__]    182
                 12000677c: addq     $28,$29,$28        [______]    183
0000000000000000 120006780: jmp      $31 ($28),0x000000 [___A_C]    184

other:

                 12002f210: subl     $14,0x01,$16       [____B_]  29028
                 12002f214: ldl      $8 ,0x0058($30)    [___A__]  29029
000000000000010b 12002f218: lda      $18,0x010b($31)    [___ABC]  29030
0000000120000d80 12002f21c: ldq      $28,0x90b0($29)    [___ABC]  29031
                 12002f220: cmpult   $18,$16,$18        [___A__]  29032
                 12002f228: s8addq   $8 ,$12,$13        [______]  29034
0000000000000000 12002f22c: stl      $8 ,0x00a8($30)    [_____C]  29035
0000000000000000 12002f230: stq      $13,0x00b0($30)    [_____C]  29036
                 12002f234: s4addq   $16,$28,$28        [____B_]  29037
0000000000000000 12002f238: bne      $18,bad            [____BC]  29038

                 12002f23c: lda      $17,0x0000($8)     [_M_A__]  29039
                 12002f240: ldl      $28,0x0054($28)    [___A__]  29040
                 12002f244: addq     $28,$29,$28        [______]  29041
0000000000000000 12002f248: jmp      $31 ($28),0x000000 [___A_C]  29042


=======================================================*/


#undef __PROC__
#define __PROC__ "AinsIsSignExtension"

static BOOL AinsIsSignExtension(INDEX iins,REG reg)
{
    if( AINS_CODE(iins) == I_zapnot )
    {
        if( AINS_REG_B(iins) == REG_LIT && AINS_DATA(iins) == 0xf && AINS_REG_A(iins) == reg ) return 1;
        else return 0;
    }
    else if( AINS_CODE(iins) == I_zap )
    {
        if( AINS_REG_B(iins) == REG_LIT && AINS_DATA(iins) == 0xf0 && AINS_REG_A(iins) == reg ) return 1;
        else return 0;
    }
    else if( AINS_CODE(iins) == I_addl )
    {
        if( AINS_REG_B(iins) == REG_ZERO && AINS_REG_A(iins) == reg ) return 1;
        else if( AINS_REG_A(iins) == REG_ZERO && AINS_REG_B(iins) == reg ) return 1;
        else return 0;
    }
    else
        return 0;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindMaxValue"

NUMBER FindMaxValue(INDEX ibbl, INDEX prev, REG reg)
{
    INDEX iins_ibr;
    INDEX iins_def;
    INDEX iins_se;
    REG   the_reg;
    BOOL  swap;
    BOOL  cond;
    
    iins_ibr = ABBL_INS_LAST(prev);

    CASSERT( AINS_TYPE(iins_ibr) == IT_IBR,("Expected it_ibr type instruction\n"));
    if( AEDG_TAIL(ABBL_EDG_TRUE(prev)) == ibbl )
    {
        cond = 1;
    }
    else
    {
        CASSERT( AEDG_TAIL(ABBL_EDG_FALSE(prev)) == ibbl,("unexpected bbl inconsistency\n"));
        cond = 0;
    }
    
    the_reg = AINS_REG_A(iins_ibr);
    
    for( iins_def=AINS_PREV(iins_ibr); VALID(iins_def); iins_def=AINS_PREV(iins_def))
    {
        if( AINS_REG_C(iins_def) == the_reg ) break;
    }
    
    if( !VALID(iins_def) ) return INVALID;
    
    if( AINS_CODE(iins_ibr) == I_beq || AINS_CODE(iins_ibr) == I_bne )
    {
        if(  AINS_CODE(iins_def) != I_cmple  &&
             AINS_CODE(iins_def) != I_cmplt  &&
             AINS_CODE(iins_def) != I_cmpule &&
             AINS_CODE(iins_def) != I_cmpult )
        {
            WARNING(("@@@@@@@@@@@ bad opcode\n"));
            return INVALID;
        }

           
        swap = 0;

        if( AINS_CODE(iins_ibr) == I_beq &&  cond == 1 ) swap = 1;
        if( AINS_CODE(iins_ibr) == I_bne &&  cond == 0 ) swap = 1;
        

        if( swap == 0 )
        {
            if( !(AINS_FLAGS(iins_def) & FI_EVAL_B) )
            {
                WARNING(("regb of iins %s not evaluated\n",AINS_NAME(iins_def)));
                return INVALID;
            }

            if( reg != AINS_REG_A(iins_def) )
            {
                for( iins_se=AINS_PREV(iins_def); VALID(iins_se); iins_se=AINS_PREV(iins_se))
                {
                    if( AINS_REG_C(iins_se) == AINS_REG_A(iins_def) )
                        break;
                }

                if( !AinsIsSignExtension(iins_se,reg) )
                {
                    WARNING(("cannot find register definition\n"));
                    return INVALID;
                }
                
            }
                    
            switch( AINS_CODE(iins_def) )
            {
            case I_cmple:
            case I_cmpule:                       
                return AINS_B_EVAL(iins_def);
            case I_cmplt:
            case I_cmpult:                       
                return AINS_B_EVAL(iins_def)-1;
            }
        }
        else
        {
            if( !(AINS_FLAGS(iins_def) & FI_EVAL_A) )
            {
                WARNING(("rega of iins %s not evaluated\n",AINS_NAME(iins_def)));
                return INVALID;
            }

            if( reg != AINS_REG_B(iins_def) )
            {
                for( iins_se=AINS_PREV(iins_def); VALID(iins_se); iins_se=AINS_PREV(iins_se))
                {
                    if( AINS_REG_C(iins_se) == AINS_REG_B(iins_def) )
                        break;
                }

                if( !AinsIsSignExtension(iins_se,reg) )
                {
                    WARNING(("cannot find register\n"));                    
                    return INVALID;
                }
                
            }
                        
            switch( AINS_CODE(iins_def) )
            {
            case I_cmple:
            case I_cmpule:                       
                return AINS_A_EVAL(iins_def)-1;
            case I_cmplt:
            case I_cmpult:                       
                return AINS_A_EVAL(iins_def);
            }
        }
    }
    else if( AINS_CODE(iins_ibr) == I_bgt ||
             AINS_CODE(iins_ibr) == I_bge ||
             AINS_CODE(iins_ibr) == I_ble ||
             AINS_CODE(iins_ibr) == I_blt )
    {
        if( AINS_CODE(iins_def) != I_lda )
        {
            WARNING(("unexpected iins_def opcode %s\n",AINS_NAME(iins_def)));
            return INVALID;
        }

        if( AINS_REG_B(iins_def) != reg )
        {
            WARNING(("cannot find register\n"));
            return INVALID;
        }
        
        switch( AINS_CODE(iins_ibr) )
        {
        case I_bge:
            if( cond )
            {
                return INVALID;
            }
            return -AINS_DATA(iins_def) -1;
        case I_bgt:
            if( cond )
            {
                return INVALID;
            }
            return -AINS_DATA(iins_def);
        case I_ble:
            if( cond==0 )
            {
                return INVALID;
            }
            
            return -AINS_DATA(iins_def);
        case I_blt:                       
            if( cond==0 )
            {
                return INVALID;
            }
            return -AINS_DATA(iins_def) -1;
        default:
            ERROR(("unexpected case\n"));
            return INVALID;
            
        }
    }

    return INVALID;
}






/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindTableJmp"

extern INDEX FindTableJmp(INDEX ibbl,INDEX prev)
{
    INDEX iins_jmp = ABBL_INS_LAST(ibbl);
    
    if( AINS_CODE(iins_jmp) != I_jmp )
    {
        INDEX ifun = ABBL_FUN(ibbl);
        WARNING(("could not find jmp for bt_jump %d in %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
        return INVALID;
    }

    return iins_jmp;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindTableAdd"

extern INDEX FindTableAdd(INDEX ibbl,INDEX prev,INDEX iins_jmp)
{
    REG reg_jmp = AINS_REG_B(iins_jmp);
    INDEX iins_add;
    
    for( iins_add=AINS_PREV(iins_jmp); VALID(iins_add); iins_add=AINS_PREV(iins_add))
    {
        if( AINS_REG_C(iins_add) == reg_jmp ) break;
    }
    
    if( !VALID(iins_add) )
        FOREACH_BBL_INS_R(prev,iins_add)
        {
            if( AINS_REG_C(iins_add) == reg_jmp ) break;
        }
    
    if( !VALID(iins_add) ||
        AINS_CODE(iins_add) != I_addq ||
        AINS_REG_B(iins_add) != REG_GP)
    {
        INDEX ifun = ABBL_FUN(ibbl);
        WARNING(("could not find add for bt_jump %d in %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
        return INVALID;
    }

    return iins_add;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindTableLdl"

extern INDEX FindTableLdl(INDEX ibbl,INDEX prev,INDEX iins_add)
{
    REG reg_offset = AINS_REG_A(iins_add);
    INDEX iins_ldl;
    
    for( iins_ldl=AINS_PREV(iins_add); VALID(iins_ldl); iins_ldl=AINS_PREV(iins_ldl))
    {
        if( AINS_REG_C(iins_ldl) == reg_offset ) break;
    }
    
    if( !VALID(iins_ldl) )
        FOREACH_BBL_INS_R(prev,iins_ldl)
        {
            if( AINS_REG_C(iins_ldl) == reg_offset ) break;
        }
    
    if( !VALID(iins_ldl) || AINS_CODE(iins_ldl) != I_ldl )
    {
        INDEX ifun = ABBL_FUN(ibbl);
        WARNING(("could not find ldl for bt_jump %d in %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
        return INVALID;
    }

    return iins_ldl;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "FindTableInd"

extern INDEX FindTableInd(INDEX ibbl,INDEX prev,INDEX iins_ldl)
{
    REG reg = AINS_REG_B(iins_ldl);
    INDEX iins_ind;
    
    for( iins_ind=AINS_PREV(iins_ldl); VALID(iins_ind); iins_ind=AINS_PREV(iins_ind))
    {
        if( AINS_REG_C(iins_ind) == reg ) break;
    }
    
    if( !VALID(iins_ind) )
        FOREACH_BBL_INS_R(prev,iins_ind)
        {
            if( AINS_REG_C(iins_ind) == reg ) break;
        }
    
    if( !VALID(iins_ind) ||
        AINS_CODE(iins_ind)!=I_s4addq ||
        !(AINS_FLAGS(iins_ind)&FI_EVAL_B) )
    {
        INDEX ifun = ABBL_FUN(ibbl);
        WARNING(("could not find ind for bt_jump %d in %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
        return INVALID;
    }

    return iins_ind;
}



/*=======================================================


=======================================================*/
#undef __PROC__
#define __PROC__ "AbblNormalizeJump2"

static void AbblNormalizeJump2(INDEX ibbl)
{
    INDEX pred;
    INDEX iins;

    REGLIST def;
    REGLIST use;
    
    INDEX iins_ind;
    INDEX iins_ldl;
    INDEX iins_add;   
    INDEX iins_jmp;
    
    CASSERT( ABBL_TYPE(ibbl) == BT_JUMP,("non bt_jump bbl\n"));
    CASSERT( !VALID(ABBL_JUMP_TAB(ibbl)),("valid jump_tab bbl\n"));

    if( AbblNpreds(ibbl) != 1 ) return;
    pred = AEDG_HEAD( ABBL_PRED(ibbl) );
    
    /* find JMP instructions =======================================================*/
    
    iins_jmp = ABBL_INS_LAST(ibbl);
    if( !VALID(iins_jmp) ) return;

    /* find ADD instructions =======================================================*/
    
    for( iins_add=AINS_PREV(iins_jmp); VALID(iins_add); iins_add=AINS_PREV(iins_add))
    {
        if( AINS_REG_C(iins_add) == AINS_REG_B(iins_jmp)) break;
    }
    
    if( !VALID(iins_add) )
    {
        def = REGS_NONE;
        use = REGS_NONE;
        
        FOREACH_BBL_INS_R(pred,iins_add)
        {
            if( AINS_REG_C(iins_add) == AINS_REG_B(iins_jmp) ) break;
            def |= AinsDef(iins_add);
            use |= AinsUse(iins_add);
        }
                
        if( !VALID(iins_add) ) return;
        if( AINS_CODE(iins_add) != I_addq ) return;
        if( AINS_REG_B(iins_add) != REG_GP ) return;
        if( AinsUse(iins_add) & def )
        {
            WARNING(("def after use\n"));
            PrintBblSuper(stdout,pred);
            return;
        }
        if( AinsDef(iins_add) & use )
        {
            WARNING(("use after def\n"));
            PrintBblSuper(stdout,pred);
            return;
        }

        iins = AinsDup(iins_add);
        AINS_BBL(iins) = ibbl;
        AinsKill(iins_add);
        AinsPrepend(iins);
        iins_add = iins;
    }
    else
    {
        if( AINS_CODE(iins_add) != I_addq ) return;
        if( AINS_REG_B(iins_add) != REG_GP ) return;
    }
    
    /* find LDL instructions =======================================================*/
    
    for( iins_ldl=AINS_PREV(iins_add); VALID(iins_ldl); iins_ldl=AINS_PREV(iins_ldl))
    {
        if( AINS_REG_C(iins_ldl) == AINS_REG_A(iins_add) ) break;
    }

    if( !VALID(iins_ldl) )
    {
        def = REGS_NONE;
        use = REGS_NONE;
        
        FOREACH_BBL_INS_R(pred,iins_ldl)
        {
            if( AINS_REG_C(iins_ldl) == AINS_REG_A(iins_add) ) break;
            def |= AinsDef(iins_ldl);
            use |= AinsUse(iins_ldl);
        }
                
        if( !VALID(iins_ldl) ) return;
        if( AINS_CODE(iins_ldl) != I_ldl ) return;
        if( AinsUse(iins_ldl) & def )
        {
            WARNING(("def after use\n"));
            PrintBblSuper(stdout,pred);
            return;
        }
        if( AinsDef(iins_ldl) & use )
        {
            WARNING(("use after def\n"));
            PrintBblSuper(stdout,pred);
            return;
        }

        iins = AinsDup(iins_ldl);
        AINS_BBL(iins) = ibbl;
        AinsKill(iins_ldl);
        AinsPrepend(iins);
        iins_ldl = iins;
    }
    else
    {
        if( AINS_CODE(iins_ldl) != I_ldl ) return;
    }
        
    /* find IND instructions =======================================================*/
    
    for( iins_ind=AINS_PREV(iins_ldl); VALID(iins_ind); iins_ind=AINS_PREV(iins_ind))
    {
        if( AINS_REG_C(iins_ind) == AINS_REG_B(iins_ldl) ) break;
    }
    if( !VALID(iins_ind) )
    {
        def = REGS_NONE;
        use = REGS_NONE;
        
        FOREACH_BBL_INS_R(pred,iins_ind)
        {
            if( AINS_REG_C(iins_ind) == AINS_REG_B(iins_ldl) ) break;
            def |= AinsDef(iins_ind);
            use |= AinsUse(iins_ind);
        }
                
        if( !VALID(iins_ind) ) return;
        if( AINS_CODE(iins_ind) != I_s4addq ) return;
        if( !(AINS_FLAGS(iins_ind)&FI_EVAL_B) ) return;
        if( AinsUse(iins_ind) & def )
        {
            WARNING(("def after use\n"));
            PrintBblSuper(stdout,pred);
            return;
        }
        if( AinsDef(iins_ind) & use )
        {
            WARNING(("use after def\n"));
            PrintBblSuper(stdout,pred);
            return;
        }

        iins = AinsDup(iins_ind);
        AINS_BBL(iins) = ibbl;
        AinsKill(iins_ind);
        AinsPrepend(iins);
        iins_ind = iins;
    }
    else
    {
        if( AINS_CODE(iins_ind) != I_s4addq ) return;
        if( !(AINS_FLAGS(iins_ind)&FI_EVAL_B) ) return;
    }
}

/*=======================================================
=======================================================*/


#undef __PROC__
#define __PROC__ "AbblNormalizeJump1"

static void AbblNormalizeJump1(INDEX ibbl)
{
    INDEX pred;
    INDEX iins;
    INDEX iedg;
    
    REG   reg_src;
    REG   reg_dst;
    REG   the_reg;
    REGLIST avail;
    
    INDEX iins_lda2;
    INDEX iins_cmv2;
    INDEX iins_sra2;
    INDEX iins_cmp2;
    INDEX iins_ibr2;
    
    INDEX iins_lda;
    INDEX iins_cmv;
    INDEX iins_sra;
    INDEX iins_ind;
    INDEX iins_ldl;
    INDEX iins_add;   
    INDEX iins_jmp;
    
    CASSERT( ABBL_TYPE(ibbl) == BT_JUMP,("non bt_jump bbl\n"));
    CASSERT( !VALID(ABBL_JUMP_TAB(ibbl)),("valid jump_tab bbl\n"));

    /* find JMP instructions =======================================================*/
    
    iins_jmp = ABBL_INS_LAST(ibbl);
    if( !VALID(iins_jmp) ) return;

    /* find ADD instructions =======================================================*/
    
    for( iins_add=AINS_PREV(iins_jmp); VALID(iins_add); iins_add=AINS_PREV(iins_add))
    {
        if( AINS_REG_C(iins_add) == AINS_REG_B(iins_jmp)) break;
    }
    
    if( !VALID(iins_add) )  return;
    if( AINS_CODE(iins_add) != I_addq ) return;
    if( AINS_REG_B(iins_add) != REG_GP ) return;

    /* find LDL instructions =======================================================*/
    
    for( iins_ldl=AINS_PREV(iins_add); VALID(iins_ldl); iins_ldl=AINS_PREV(iins_ldl))
    {
        if( AINS_REG_C(iins_ldl) == AINS_REG_A(iins_add) ) break;
    }
    
    if( !VALID(iins_ldl) ) return;
    if( AINS_CODE(iins_ldl) != I_ldl ) return;

    /* find IND instructions =======================================================*/
    
    for( iins_ind=AINS_PREV(iins_ldl); VALID(iins_ind); iins_ind=AINS_PREV(iins_ind))
    {
        if( AINS_REG_C(iins_ind) == AINS_REG_B(iins_ldl) ) break;
    }

    if( !VALID(iins_ind) ) return;
    if( AINS_CODE(iins_ind) != I_s4addq ) return;
    if( !(AINS_FLAGS(iins_ind)&FI_EVAL_B) ) return;

    /* find SRA instructions =======================================================*/

    for( iins_sra=AINS_PREV(iins_ind); VALID(iins_sra); iins_sra=AINS_PREV(iins_sra))
    {
        if( AINS_REG_C(iins_sra) == AINS_REG_A(iins_ind) ) break;
    }
        
    if( !VALID(iins_sra) ) return;
    if( AINS_CODE(iins_sra) != I_sra ) return;
    if( AINS_REG_A(iins_sra) != AINS_REG_C(iins_sra) ) return;
    if( !(AINS_FLAGS(iins_sra)&FI_EVAL_B) ) return;

    /* find CMV instructions =======================================================*/

    for( iins_cmv=AINS_PREV(iins_sra); VALID(iins_cmv); iins_cmv=AINS_PREV(iins_cmv))
    {
        if( AINS_REG_C(iins_cmv) == AINS_REG_A(iins_sra) ) break;
    }
    
    if( !VALID(iins_cmv) ) return;
    if( AINS_TYPE(iins_cmv) != IT_ICM ) return;

    /* find LDA instructions =======================================================*/
    
    for( iins_lda=AINS_PREV(iins_cmv); VALID(iins_lda); iins_lda=AINS_PREV(iins_lda))
    {
        if( AINS_REG_C(iins_lda) == AINS_REG_A(iins_sra) ) break;
    }

    if( !VALID(iins_lda) ) return;
    if( AINS_CODE(iins_lda) != I_lda ) return;

    /* ============== SO FAR SO GOOD =============================== */
    
    reg_dst = AINS_REG_A(iins_ind);
    reg_src = AINS_REG_B(iins_lda);

    if( reg_src != AINS_REG_A(iins_cmv) ) return;
    if( reg_src != AINS_REG_B(iins_cmv) ) return;
    if( reg_src != AINS_REG_B(iins_lda) ) return;

    for( iins=AINS_PREV(iins_lda); VALID(iins); iins=AINS_PREV(iins))
    {
        if( AINS_REG_C(iins) == reg_src ) return;
        if( AINS_REG_C(iins) == reg_dst ) return;
    }

    /* check the predecessors =======================================================*/

    FOREACH_BBL_PRED(ibbl,iedg)
    {
        pred = AEDG_HEAD(iedg);
            
        /* find IBR2 instructions =======================================================*/
    
        iins_ibr2 = ABBL_INS_LAST(pred);

        if( !VALID(iins_ibr2) ) return;
        if( AINS_CODE(iins_ibr2) != I_beq && AINS_CODE(iins_ibr2) != I_bne ) return;

        /* find CMP2 instructions =======================================================*/
        
        for( iins_cmp2 =AINS_PREV(iins_ibr2); VALID(iins_cmp2); iins_cmp2=AINS_PREV(iins_cmp2))
        {
            if( AINS_REG_C(iins_cmp2) == AINS_REG_A(iins_ibr2) ) break;
        }
        
        if( !VALID(iins_cmp2) ) return;
        if( AINS_CODE(iins_cmp2) != I_cmpult &&
            AINS_CODE(iins_cmp2) != I_cmpule &&
            AINS_CODE(iins_cmp2) != I_cmplt &&
            AINS_CODE(iins_cmp2) != I_cmple ) return;
        if( AINS_FLAGS(iins_cmp2) & FI_EVAL_B )
        {
            the_reg = AINS_REG_A(iins_cmp2);
        }
        else if( AINS_FLAGS(iins_cmp2) & FI_EVAL_A )
        {
            the_reg = AINS_REG_B(iins_cmp2);
        }
        else
        {
            return;
        }

        /* find SRA instructions =======================================================*/
        
        for( iins_sra2=AINS_PREV(iins_cmp2); VALID(iins_sra2); iins_sra2=AINS_PREV(iins_sra2))
        {
            if( AINS_REG_C(iins_sra2) == the_reg ) break;
        }
        
        if( !VALID(iins_sra2) ) return;
        if( AINS_CODE(iins_sra2) != AINS_CODE(iins_sra) ) return;
        if( AINS_REG_A(iins_sra) != AINS_REG_C(iins_sra) ) return;
        if( !(AINS_FLAGS(iins_sra2)&FI_EVAL_B) ) return;
        if( AINS_B_EVAL(iins_sra2) != AINS_B_EVAL(iins_sra) ) return;

        /* find CMV instructions =======================================================*/

        for( iins_cmv2=AINS_PREV(iins_sra2); VALID(iins_cmv2); iins_cmv2=AINS_PREV(iins_cmv2))
        {
            if( AINS_REG_C(iins_cmv2) == AINS_REG_A(iins_sra2) ) break;
        }
    
        if( !VALID(iins_cmv2) ) return;
        if( AINS_CODE(iins_cmv2) != AINS_CODE(iins_cmv) ) return;

        /* find LDA instructions =======================================================*/
        
        for( iins_lda2=AINS_PREV(iins_cmv2); VALID(iins_lda2); iins_lda2=AINS_PREV(iins_lda2))
        {
            if( AINS_REG_C(iins_lda2) == AINS_REG_A(iins_sra2) ) break;
        }
        
        if( !VALID(iins_lda) ) return;
        if( AINS_CODE(iins_lda2) != AINS_CODE(iins_lda) ) return;
        if( AINS_DATA(iins_lda2) != AINS_DATA(iins_lda) ) return;

        /* do check  =======================================================*/

        if( reg_src != AINS_REG_A(iins_cmv2) ) return;
        if( reg_src != AINS_REG_B(iins_cmv2) ) return;
        if( reg_src != AINS_REG_B(iins_lda2) ) return;

        avail = ~AbblRegsLiveIn(ibbl);
        avail &= ~REG2BIT(reg_src);
        avail &= ~REG2BIT(reg_dst);
        avail &= ~REGS_FLT;
        avail &= ~B_ZERO;

#if 0
        printf("avail %016lx\n\n",avail);
#endif   
        if(  avail != REGS_NONE )
        {
            REG reg = RegsGetFirst(avail);
            AINS_REG_A(iins_ibr2) = reg;
            AINS_REG_C(iins_cmp2) = reg;
        }
        
        for( iins=iins_ibr2; iins!=iins_lda2; iins=AINS_PREV(iins))
        {
            if( iins == iins_sra2 ) continue;
            if( iins == iins_cmv2 ) continue;
            
            if( AINS_REG_C(iins) == reg_src ) return;
            if( AINS_REG_C(iins) == reg_dst ) return;
        }
    }

    /* transform the predecessors =======================================================*/

    FOREACH_BBL_PRED(ibbl,iedg)
    {
        pred = AEDG_HEAD(iedg);
            
        /* find IBR2 instructions =======================================================*/
    
        iins_ibr2 = ABBL_INS_LAST(pred);

        /* find CMP2 instructions =======================================================*/
        
        for( iins_cmp2 =AINS_PREV(iins_ibr2); VALID(iins_cmp2); iins_cmp2=AINS_PREV(iins_cmp2))
        {
            if( AINS_REG_C(iins_cmp2) == AINS_REG_A(iins_ibr2) ) break;
        }
        
        /* find SRA instructions =======================================================*/
        
        for( iins_sra2=AINS_PREV(iins_cmp2); VALID(iins_sra2); iins_sra2=AINS_PREV(iins_sra2))
        {
            if( AINS_REG_C(iins_sra2) == AINS_REG_A(iins_cmp2) ||
                AINS_REG_C(iins_sra2) == AINS_REG_B(iins_cmp2) ) break;
        }
        
        /* find CMV instructions =======================================================*/

        for( iins_cmv2=AINS_PREV(iins_sra2); VALID(iins_cmv2); iins_cmv2=AINS_PREV(iins_cmv2))
        {
            if( AINS_REG_C(iins_cmv2) == AINS_REG_A(iins_sra2) ) break;
        }
    
        /* find LDA instructions =======================================================*/
        
        for( iins_lda2=AINS_PREV(iins_cmv2); VALID(iins_lda2); iins_lda2=AINS_PREV(iins_lda2))
        {
            if( AINS_REG_C(iins_lda2) == AINS_REG_A(iins_sra2) ) break;
        }

        if( AINS_REG_C(iins_sra2) == AINS_REG_A(iins_cmp2) )
            AINS_REG_A(iins_cmp2) = reg_dst;
        if( AINS_REG_C(iins_sra2) == AINS_REG_B(iins_cmp2) )
            AINS_REG_B(iins_cmp2) = reg_dst;
        
        AINS_REG_C(iins_lda2) = reg_dst;
        AINS_REG_C(iins_cmv2) = reg_dst;
        AINS_REG_A(iins_sra2) = reg_dst;
        AINS_REG_C(iins_sra2) = reg_dst;

    }
    
    AinsKill(iins_lda);
    AinsKill(iins_cmv);
    AinsKill(iins_sra);

#if 0
    FOREACH_BBL_PRED(ibbl,iedg)
    {
        pred = AEDG_HEAD(iedg);
        
        PrintBblSuper(stdout,pred);
    }
    
    PrintBblSuper(stdout,ibbl);
#endif
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblCreateJumpTable"

static void AbblCreateJumpTable(INDEX ibbl)
{
    INDEX iins_adj;
    INDEX iins_ind;
    INDEX iins_ldl;
    INDEX iins_add;   
    INDEX iins_jmp;

    REG   the_reg;

    INDEX iedg;
    
    NUMBER correction;
    NUMBER entries;
    NUMBER i;

    INDEX itab;
    INDEX jtab;
    INDEX jbbl;
    
    INDEX isec;
    ADDRESS table;
    INDEX isec_text;


    CASSERT( ABBL_TYPE(ibbl) == BT_JUMP,("non bt_jump bbl\n"));
    CASSERT( !VALID(ABBL_JUMP_TAB(ibbl)),("valid jump_tab bbl\n"));


    AbblNormalizeJump2(ibbl);
    AbblNormalizeJump1(ibbl);


    /* find JMP instructions =======================================================*/
    
    iins_jmp = ABBL_INS_LAST(ibbl);
    if( !VALID(iins_jmp) ) return;

    /* find ADD instructions =======================================================*/
    
    for( iins_add=AINS_PREV(iins_jmp); VALID(iins_add); iins_add=AINS_PREV(iins_add))
    {
        if( AINS_REG_C(iins_add) == AINS_REG_B(iins_jmp)) break;
    }
    
    if( !VALID(iins_add) )  return;
    if( AINS_CODE(iins_add) != I_addq ) return;
    if( AINS_REG_B(iins_add) != REG_GP ) return;

    /* find LDL instructions =======================================================*/
    
    for( iins_ldl=AINS_PREV(iins_add); VALID(iins_ldl); iins_ldl=AINS_PREV(iins_ldl))
    {
        if( AINS_REG_C(iins_ldl) == AINS_REG_A(iins_add) ) break;
    }
    
    if( !VALID(iins_ldl) ) return;
    if( AINS_CODE(iins_ldl) != I_ldl ) return;

    /* find IND instructions =======================================================*/
    
    for( iins_ind=AINS_PREV(iins_ldl); VALID(iins_ind); iins_ind=AINS_PREV(iins_ind))
    {
        if( AINS_REG_C(iins_ind) == AINS_REG_B(iins_ldl) ) break;
    }

    if( !VALID(iins_ind) ) return;
    if( AINS_CODE(iins_ind) != I_s4addq ) return;
    if( !(AINS_FLAGS(iins_ind)&FI_EVAL_B) ) return;

    /* find ADJ instructions =======================================================*/
    for( iins_adj=AINS_PREV(iins_ind); VALID(iins_adj); iins_adj=AINS_PREV(iins_adj))
    {
        if( AINS_REG_C(iins_adj) == AINS_REG_A(iins_ind) ) break;
    }
    
    /* =======================================================*/

    the_reg = AINS_REG_A(iins_ind);
    correction = 0;
    
    if( VALID(iins_adj) )
    {
        switch( AINS_CODE(iins_adj) )
        {
        case I_lda:
            correction = AINS_DATA(iins_adj);
            the_reg = AINS_REG_B(iins_adj);
            break;
        case I_subl:
            if( !(AINS_FLAGS(iins_adj) * FI_EVAL_B) )
                return;
            correction = -AINS_B_EVAL(iins_adj);
            the_reg = AINS_REG_A(iins_adj);
            break;
        case I_zapnot:
            if( !(AINS_FLAGS(iins_adj) * FI_EVAL_B) )
                return;
            if( AINS_B_EVAL(iins_adj) != 0xf )
                return;
            correction = 0;
            the_reg = AINS_REG_A(iins_adj);
            break;
        default:
            WARNING(("bad adjust instruction: %s\n",AINS_NAME(iins_adj)));
#if 0            
            PrintIns(stdout,iins_adj);
#endif
            return;
        }
    }
    
    entries = INVALID;
    
    FOREACH_BBL_PRED(ibbl,iedg)
    {
        INDEX pred = AEDG_HEAD(iedg);
        if( ABBL_TYPE(pred) != BT_BRANCH )
        {
            WARNING(("expected all preds of %d to be BT_BRANCH, %d is not\n",ibbl,pred));
#if 0
            PrintBblSuper(stdout,pred);
            PrintBblSuper(stdout,ibbl);
#endif
            return;
        }
        
        if( !VALID(entries) )
        {
            entries = FindMaxValue(ibbl,pred,the_reg);
        }
        else
        {
            if( entries != FindMaxValue(ibbl,pred,the_reg) )
            {
                WARNING(("entry missmatch\n"));
                return;
            }
        }
        if( !VALID(entries) )
        {
            WARNING(("cannot determine entries\n"));
            return;
        }

    }

    if( !VALID(entries) )
    {
        WARNING(("cannot determine entries\n"));
        return;
    }
    
    entries += 1 + correction;
    
    if( entries <= 0 )
    {
        WARNING(("no entries\n"));
        return;
    }
    
    /* find table start =======================================================*/

    table = AINS_B_EVAL(iins_ind);
    table += AINS_DATA(iins_ldl);
    isec = LocateSection(table);

    if( !AsecIsReadOnly(isec) )
    {
        WARNING(("strange section at address %016lx for bt_jump %d  \n",table,ibbl));
#if 0
        PrintSec(stdout,isec);
#endif
        return;
    }

    CERROR( AsecIsCode(isec),("read from code segment\n"));
            
    /* find number of entries  =======================================================*/
    
#if 0
    printf("@@@@@@@@@@@@@@@@@@@@@@@ENTRIES: %d (correction %d)\n",entries,correction);
#endif
    isec_text = SysGetTextSection();

    itab = INVALID;

#if 0
    PrintSep1(stdout);
    printf("entries: %d table start %016lx\n",entries,table);
    if( VALID(iins_zap) )
        PrintIns(stdout,iins_zap);
    PrintIns(stdout,iins_cmp);
    PrintIns(stdout,iins_ibr);
    if( VALID(iins_adj) )
        PrintIns(stdout,iins_adj);
    PrintIns(stdout,iins_ind);
    PrintIns(stdout,iins_ldl);
    PrintIns(stdout,iins_add);
    PrintIns(stdout,iins_jmp);
    PrintBblSuper(stdout,prev);
    PrintBblSuper(stdout,ibbl);
    
#endif

    for(i=0;i<entries;i++)
    {
        ADDRESS target = SysGetGlobalPointer() + ReadData32(table+i*4);
    
        if( !AsecContainsAddress(isec_text,target))
        {
            WARNING(("bad target address: %016lx\n",target));
#if 0
            PrintBblSuper(stdout,ibbl);
#endif
            return;
        }
        
        jbbl = FindTargetBbl(target);
        jtab = AtabNew();
        ATAB_TARGET(jtab) = jbbl;
        if( !VALID(itab) )
            itab = jtab;
#if 0        
        printf("%5d %16lx %5d\n",ibbl,target,jbbl);
#endif        
        CASSERT( ABBL_FLAGS(jbbl) & F_GPREL32,("jump target is not gprel32\n"));
                
        if( ABBL_FUN(jbbl) != ABBL_FUN(ibbl) )
        {
            WARNING(("jump to different function: %016lx\n",target));
            return;
        }
        
    }

    /* mark end of table */
    
    jtab = AtabNew();
    ATAB_TARGET(jtab) = INVALID;
    ABBL_JUMP_TAB(ibbl) = itab;
}
/*=======================================================
@OPT
Identify possible targets in a BT_JUMP basic block.
If all targets can be identifies the block changes its type to BT_SWITCH
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptTableJumps"

#define DEBUG_SWITCH 1

extern BOOL OptTableJumps(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Switch");
    BOOL   change;
    INDEX  ifun;
    NUMBER jcount = 0;
    NUMBER ecount = 0;
    NUMBER pcount = 0;
    INDEX  ibbl;
    INDEX  jbbl;
    INDEX  itab;
    INDEX  iedg;
    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_JUMP ) continue;
        CASSERT( !VALID(ABBL_JUMP_TAB(ibbl)),("jump table not valid\n"));

        ifun = ABBL_FUN(ibbl);
        
        if( verbose )
            fprintf(fp,"\t[*] jump in bbl: %d  fun: %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun));

        AbblCreateJumpTable(ibbl);
        
        pcount++;

        if( !VALID(ABBL_JUMP_TAB(ibbl)) )
        {
            if( verbose )
            {
                FOREACH_BBL_PRED(ibbl,iedg)
                {
                    PrintBblSuper(stdout,AEDG_HEAD(iedg));
                }
                
                PrintBblSuper(stdout,ibbl);
            }
            
            continue;
        }
        

        CASSERT( VALID(ABBL_EDG_JUMP(ibbl)),("weird\n") );
        
        if( verbose )
        {
            fprintf(fp,"del: ");
            PrintEdg(fp,ABBL_EDG_JUMP(ibbl));
        }
        
        AedgKillFancy(ABBL_EDG_JUMP(ibbl));
        ABBL_TYPE(ibbl) = BT_SWITCH;
        jcount++;

        FOREACH_BBL_TAB(ibbl,itab)
        {
            jbbl = ATAB_TARGET(itab);

            /* do not add duplicate edges */

            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_TAIL(iedg) == jbbl )
                    break;
            }
            
            if( VALID(iedg) ) continue;

            /* create a new edge */
            
            iedg = AedgNewFancy(ibbl,jbbl,ET_SWITCH);
            
            if( verbose )
            {
                fprintf(fp,"add: ");
                PrintEdg(fp,iedg);
            }
                    
            AedgLinkFancy( iedg);
            ecount++;

            CWARNING( AEDG_IS_INTER(iedg),("created interproc et_switch edge\n"));            
        }

        AbblInstallNewJumpTable(ibbl);
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = jcount>0;
    InfoStatus(fp,"\t[OptSwitch%c] Resolved Jumps: %d (out of %d) Created Edgs: %d {%d ms}\n",
            CHANGE(change),jcount,pcount,ecount,time);
    return change;
}


/*=======================================================
@OPT
If all BT_JUMP block in a function have been converted into
BT_SWITCH blocks. We feel it is safe to eliminate
edges from HELL to GPREL32 basic blocks in this function.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptGprel"


extern BOOL OptGprel(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Gprel");
    BOOL   change;
    INDEX  ifun;
    NUMBER count = 0;
    INDEX  ibbl;
    INDEX iedg;
    clock_t time = clock();
    
    FOREACH_LIVE_FUN(ifun)
    {
        if( AfunNJumps(ifun) > 0 ) continue;
        if( AfunNSwitches(ifun) == 0 ) continue;

        FOREACH_FUN_BBL(ifun,ibbl)
         {
             if( !(ABBL_FLAGS(ibbl) & F_GPREL32) ) continue;

             FOREACH_BBL_PRED(ibbl,iedg)
             {
                 if( AEDG_TYPE(iedg) == ET_HELL )
                 {
#if 0
                     PrintEdgSuper(stdout,iedg);
                     PrintBblSuper(stdout,AEDG_HEAD(iedg));
                     PrintBblSuper(stdout,AEDG_TAIL(iedg));
#endif                
                     CWARNING( AbblNpreds(ibbl)==1,("deleting only pred of bbl: %d %d [%s]\n",ibbl,ifun,AFUN_NAME(ifun)));
                     if( verbose )
                     {
                         fprintf(fp,"\t[*] del edge in function %d [%s]",ifun,AFUN_NAME(ifun));
                         PrintEdg(fp,iedg);
                     }
                 
                     AedgKillFancy(iedg);
                     count++;
                 }
             }
         }
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = count>0;
    InfoStatus(fp,"\t[OptGprel%c] Removed Edgs: %d {%d ms}\n",CHANGE(change),count,time);
    return change;
}


/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblGetRelocIns"

extern INDEX AbblGetRelocIns(INDEX ibbl)
{
    INDEX iins;
    CASSERT( ABBL_FLAGS(ibbl) & F_RELOC,("weird bbl\n"));
    
    FOREACH_INS(iins)
    {
        if( !(AINS_FLAGS(iins) & F_RELOC) ) continue;
        if( AINS_BBL(iins) != ibbl ) continue;
        return iins;
    }

    ERROR(("weird\n"));
    return INVALID;
}

/*=======================================================
@OPT
Given a basic block of type BT_SWITCH generate a new jumptable
for it in the enlarged data section.
This useful when we duplicated this basic block
and now need to use a different jumptable than the original.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblInstallNewJumpTable"

extern void AbblInstallNewJumpTable(INDEX ibbl)
{
    INDEX isec_data = SysGetSwitchSection();
    INDEX itab;
    INDEX jins;
    INDEX jbbl;
    INDEX prev;
    ADDRESS table;
    INDEX iins_jmp;
    INDEX iins_add;
    INDEX iins_ldl;
    INDEX iins_ind;
    NUMBER high;
    NUMBER low;
    
    CASSERT( ABBL_TYPE(ibbl) == BT_SWITCH,("non bt_switch bbl: %d\n",ibbl));
    CASSERT( VALID(ABBL_JUMP_TAB(ibbl)),("invalid jump_tab fo bbl: %d\n",ibbl));

    /* fix up the entries */

    FOREACH_BBL_TAB(ibbl,itab)
    {
        jbbl = ATAB_TARGET(itab);
        jins = AbblGetRelocIns(jbbl);
        CASSERT( ABBL_FLAGS(jbbl) & F_GPREL32,("target bbl %d is not F_GPREL32\n",jbbl));
        AINS_FLAGS(jins) |= F_GPREL32;
#if 0
        printf("bbl: %d iins: %d gprel32: %8x\n",jbbl,jins,Ins2Addr(jins)-GetGlobalPointer());
#endif

#if 1
        AsecWriteData32(isec_data,itab,Ins2Addr(jins) - SysGetGlobalPointer() );
#else
        ASEC_DATA32(isec_data)[itab] = Ins2Addr(jins) - SysGetGlobalPointer();
#endif        
    }

    /* last entry is invalid */

    CASSERT( ATAB_TARGET(itab) == INVALID,("last table entry is valid\n"));

#if 1
    AsecWriteData32(isec_data,itab,INVALID );
#else
    ASEC_DATA32(isec_data)[itab] = INVALID;
#endif
    
    /* patch up address */

    
    table = AsecGetAddressData32(isec_data, ABBL_JUMP_TAB(ibbl));

#if 0
    CASSERT( AbblNpreds(ibbl) == 1,(" num of preds != 1\n"));
#endif
    
    prev = AEDG_HEAD( ABBL_PRED(ibbl) );


    /* find JMP instructions =======================================================*/

    iins_jmp = FindTableJmp(ibbl,prev);
    if( !VALID(iins_jmp) )
    {
        ERROR(("invalid iins_jmp\n"));
        PrintBblSuper(stdout,prev);
        PrintBblSuper(stdout,ibbl);
    }
    
    /* find ADD instructions =======================================================*/

    iins_add = FindTableAdd(ibbl,prev,iins_jmp);
    if( !VALID(iins_add) )
    {
        ERROR(("invalid iins_add\n"));
        PrintBblSuper(stdout,prev);
        PrintBblSuper(stdout,ibbl);
    }
        
    /* find LDL instructions =======================================================*/

    iins_ldl = FindTableLdl(ibbl,prev,iins_add);
    if( !VALID(iins_ldl) )
    {
        ERROR(("invalid iins_ldl\n"));
        PrintBblSuper(stdout,prev);
        PrintBblSuper(stdout,ibbl);    }
    
    
    /* find IND instructions =======================================================*/

    iins_ind = FindTableInd(ibbl,prev,iins_ldl);
    if( !VALID(iins_ind) )
    {
        ERROR(("invalid iins_ind\n"));
        PrintBblSuper(stdout,prev);
        PrintBblSuper(stdout,ibbl);
    }
    
            
    /* insert instructions to load the new table address */

#if 0
    WARNING(("installing a new table at %16lx for bbl %d  itab: %d\n",table,ibbl,itab));
#endif
    
    table -= SysGetGlobalPointer();
    AINS_DATA(iins_ldl) = 0;
    AINS_FLAGS(iins_ind) &= ~FI_EVAL_B;


    high = table >> 16;
    low = table & 0xffff;
    if( low & 0x8000 )
    {
        low |= 0xffff0000;
        high += 1;
    }
#if 0
    WARNING(("table %16lx  high %8lx  low %8x\n",table,high,low));

#endif
    CASSERT( ((long)high << 16) + (long) low == table,("bad table address computation\n"));
    
    jins = AinsNew();
    AinsMakeLdah(jins,REG_GP,AINS_REG_B(iins_ind),high);
    AINS_BBL(jins) = AINS_BBL(iins_ind);
    AinsInsertBefore(jins,iins_ind);

    jins = AinsNew();
    AinsMakeLda(jins,AINS_REG_B(iins_ind),AINS_REG_B(iins_ind),low);    
    AINS_BBL(jins) = AINS_BBL(iins_ind);
    AinsInsertBefore(jins,iins_ind);

    /*
    PrintBblSuper(stdout,AINS_BBL(jins));
    */
    
}

/*=======================================================
=======================================================*/

















